package com.faraphel.tasks_valider.connectivity.task

import com.faraphel.tasks_valider.connectivity.task.session.TaskPermission
import com.faraphel.tasks_valider.connectivity.task.session.TaskSession
import com.faraphel.tasks_valider.database.TaskDatabase
import com.google.gson.Gson
import fi.iki.elonen.NanoHTTPD
import java.util.*


/**
 * A server to handle the task API to allow clients to interact with the database.
 * @param port the port of the server
 * @param database the database to interact with
 */
class TaskServer(
    private val port: Int,
    private val database: TaskDatabase
) : NanoHTTPD(port) {

    // Session

    // TODO(Faraphel): theses functions should be moved somewhere else for specific session management

    private val sessions = mutableMapOf<String, TaskSession>()  ///< sessions specific data

    /**
     * Create a new session
     * @param session the data for the session (optional)
     * @return a new session identifiant
     */
    private fun newSessionData(session: TaskSession = TaskSession()): String {
        val sessionId = UUID.randomUUID().toString()
        this.sessions[sessionId] = session
        return sessionId
    }

    /**
     * Get data from a http session
     * @param httpSession the HTTP session
     * @return the session data
     */
    private fun getSessionData(httpSession: IHTTPSession): TaskSession? {
        val sessionId = httpSession.cookies.read("sessionId") ?: return null
        val sessionData = this.getSessionData(sessionId)
        return sessionData
    }

    /**
     * Get data from a session identifiant
     * @param sessionId the identifiant of the session
     * @return the session data
     */
    private fun getSessionData(sessionId: String): TaskSession? {
        return this.sessions[sessionId]
    }

    /**
     * Get data from a http session. If it does not exist, create it.
     * @param httpSession the HTTP session
     */
    private fun getOrCreateSessionData(httpSession: IHTTPSession): TaskSession {
        // try to get the session directly
        var session = this.getSessionData(httpSession)

        // if the session does not exist, create it
        if (session == null) {
            val sessionId = this.newSessionData()
            session = this.getSessionData(sessionId)!!
        }

        // return the session
        return session
    }

    /**
     * Insert new cookies for a session in a response
     * @param response the response to inject cookies into
     * @param cookies the cookie handler
     */
    private fun responseSetSessionData(response: Response, cookies: CookieHandler): Response {
        cookies.set(Cookie("sessionId", this.newSessionData()))
        cookies.unloadQueue(response)
        return response
    }

    // HTTP

    override fun serve(httpSession: IHTTPSession): Response {
        // get the session data of the client
        val taskSession = this.getOrCreateSessionData(httpSession)

        // get the method used
        val method: Method = httpSession.method

        // parse the url
        val uri: String = httpSession.uri.substring(1)  // remove the first slash
        val uriComponents = uri.split("/")

        // get the type of the request from the uri
        val requestType = uriComponents.getOrNull(0)
            ?: return newFixedLengthResponse(
                NanoHTTPD.Response.Status.BAD_REQUEST,
                "text/plain",
                "Invalid request type"
            )

        // get the response from the correct part of the application
        val response = when (requestType) {
            // session requests
            "sessions" -> {
                TODO("Faraphel: handle admin changing permission")
            }

            // entities requests
            "entities" -> {
                val entityName = uriComponents.getOrNull(1)
                    ?: return newFixedLengthResponse(
                        NanoHTTPD.Response.Status.BAD_REQUEST,
                        "text/plain",
                        "Invalid entity name"
                    )

                this.database.handleApiRequest(taskSession, httpSession, method, entityName)
            }
            // invalid requests
            else -> {
                newFixedLengthResponse(
                    NanoHTTPD.Response.Status.BAD_REQUEST,
                    "text/plain",
                    "Invalid request type"
                )
            }
        }

        // wrap additional information in the response
        return this.responseSetSessionData(response, httpSession.cookies)
    }

    /**
     * Start the server with the default configuration
     */
    override fun start() {
        super.start(SOCKET_READ_TIMEOUT, false)
    }
}
