package com.faraphel.tasks_valider.database

import androidx.room.Database
import androidx.room.RoomDatabase
import androidx.room.TypeConverters
import com.faraphel.tasks_valider.connectivity.task.session.TaskPermission
import com.faraphel.tasks_valider.connectivity.task.session.TaskSession
import com.faraphel.tasks_valider.database.api.*
import com.faraphel.tasks_valider.database.api.base.BaseApi
import com.faraphel.tasks_valider.database.converters.InstantConverter
import com.faraphel.tasks_valider.database.dao.GroupDao
import com.faraphel.tasks_valider.database.dao.GroupStudentDao
import com.faraphel.tasks_valider.database.dao.StudentDao
import com.faraphel.tasks_valider.database.dao.TaskDao
import com.faraphel.tasks_valider.database.dao.TaskGroupDao
import com.faraphel.tasks_valider.database.dao.TeacherDao
import com.faraphel.tasks_valider.database.entities.GroupEntity
import com.faraphel.tasks_valider.database.entities.GroupStudentEntity
import com.faraphel.tasks_valider.database.entities.StudentEntity
import com.faraphel.tasks_valider.database.entities.TaskEntity
import com.faraphel.tasks_valider.database.entities.TaskGroupEntity
import com.faraphel.tasks_valider.database.entities.TeacherEntity
import fi.iki.elonen.NanoHTTPD


/**
 * The database for the tasks application.
 * Contains the entities and the relations between them.
 */
@Database(
    entities = [
        GroupEntity::class,
        StudentEntity::class,
        TeacherEntity::class,
        TaskEntity::class,

        GroupStudentEntity::class,
        TaskGroupEntity::class,
    ],
    version = 1
)
@TypeConverters(
    InstantConverter::class
)
abstract class TaskDatabase : RoomDatabase() {
    // entities

    abstract fun groupDao(): GroupDao
    abstract fun studentDao(): StudentDao
    abstract fun teacherDao(): TeacherDao
    abstract fun taskDao(): TaskDao

    abstract fun groupStudentDao(): GroupStudentDao
    abstract fun taskGroupDao(): TaskGroupDao

    // api

    private val api: Map<String, BaseApi> = mapOf(
        "group" to GroupApi(this.groupDao()),
        "student" to StudentApi(this.studentDao()),
        "teacher" to TeacherApi(this.teacherDao()),
        "task" to TaskApi(this.taskDao()),

        "group_student" to GroupStudentApi(this.groupStudentDao()),
        "task_group" to TaskGroupApi(this.taskGroupDao()),
    )

    /**
     * dispatch an API request
     */
    fun handleApiRequest(
        taskSession: TaskSession,
        httpSession: NanoHTTPD.IHTTPSession,
        method: NanoHTTPD.Method,
        entityName: String
    ): NanoHTTPD.Response {
        // get the correspond Api object for this entity
        val entityApi = this.api[entityName]
            ?: return NanoHTTPD.newFixedLengthResponse(
                NanoHTTPD.Response.Status.NOT_FOUND,
                "text/plain",
                "Invalid entity name"
            )

        // dispatch the request to the correct entity API
        return when (method) {
            // check if the data is in the database
            NanoHTTPD.Method.HEAD -> {
                if (taskSession.permission < TaskPermission.STUDENT)
                    return NanoHTTPD.newFixedLengthResponse(
                        NanoHTTPD.Response.Status.FORBIDDEN,
                        "text/plain",
                        "Forbidden"
                    )

                return entityApi.head(httpSession)
            }
            // get the data from the database
            NanoHTTPD.Method.GET -> {
                if (taskSession.permission < TaskPermission.STUDENT)
                    return NanoHTTPD.newFixedLengthResponse(
                        NanoHTTPD.Response.Status.FORBIDDEN,
                        "text/plain",
                        "Forbidden"
                    )

                return entityApi.get(httpSession)
            }
            // insert the data into the database
            NanoHTTPD.Method.POST -> {
                if (taskSession.permission < TaskPermission.TEACHER)
                    return NanoHTTPD.newFixedLengthResponse(
                        NanoHTTPD.Response.Status.FORBIDDEN,
                        "text/plain",
                        "Forbidden"
                    )

                return entityApi.post(httpSession)
            }
            // delete the data from the database
            NanoHTTPD.Method.DELETE -> {
                if (taskSession.permission < TaskPermission.TEACHER)
                    return NanoHTTPD.newFixedLengthResponse(
                        NanoHTTPD.Response.Status.FORBIDDEN,
                        "text/plain",
                        "Forbidden"
                    )

                return entityApi.delete(httpSession)
            }
            // other methods are not allowed
            else ->
                return NanoHTTPD.newFixedLengthResponse(
                    NanoHTTPD.Response.Status.METHOD_NOT_ALLOWED,
                    "text/plain",
                    "Method not allowed"
                )
        }
    }
}
